/*
 * collector.c
 * WiFi Data Elements Collector main
 *
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in.h>

#include <easy/easy.h>
#include <wifiutils.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <libubus.h>

#include "timer.h"
#include <cmdu.h>
#include <1905_tlvs.h>
#include <easymesh.h>
#include <map_module.h>


#include "backhaul_topology.h"
#include "wifi_dataelements.h"
#include "decollector_i.h"
#include "decollector.h"
#include "debug.h"
#include "config.h"


const char *getter_state2str(uint8_t s)
{
#define T2STR(s)	case s: return #s;

	switch (s) {
	T2STR(GT_IDLE)
	T2STR(GT_TOPOLOGY_RESPONSE_PENDING)
	T2STR(GT_TOPOLOGY_RESPONSE_RECEIVED)
	T2STR(GT_AP_CAP_RECEIVED)
	T2STR(GT_AP_METRIC_RECEIVED)
	T2STR(GT_BSTA_CAP_RECEIVED)
	T2STR(GT_SCAN_RESPONSE_RECEIVED)
	T2STR(GT_DONE_1)
	}

	return "UNKNOWN";

#undef T2STR
}

void decollector_stop_getter_all(struct decollector_private *p)
{
	struct wifi_network *net = &p->dm->network;
	struct wifi_network_device *dev = NULL;

	list_for_each_entry(dev, &net->devicelist, list) {
		decollector_stop_getter(p, dev);
	}
}

int decollector_stop_getter(struct decollector_private *p,
		struct wifi_network_device *dev)
{
	struct getter_context *gt;

	if (!dev->priv)
		return -1;

	gt = dev->priv;
	timer_del(&gt->t);
	fprintf(stderr, "Getter for " MACFMT " stopped in state %s\n",
		MAC2STR(dev->macaddr), getter_state2str(gt->state));
	//gt->state = GT_DONE_1;
	return 0;
}

int decollector_getter_set_state(struct decollector_private *p,
		struct wifi_network_device *dev, uint8_t state)
{
	struct getter_context *gt;

	if (state > GT_DONE_1)
		return -1;

	fprintf(stdout, "Getter for " MACFMT " set to state %s\n",
			MAC2STR(dev->macaddr), getter_state2str(state));
	gt = dev->priv;
	gt->state = state;
	return 0;
}

int decollector_getter_get_state(struct decollector_private *p,
		struct wifi_network_device *dev, uint8_t *state)
{
	struct getter_context *gt;

	gt = dev->priv;
	*state = gt->state;
	return 0;
}

int decollector_getter_running(struct decollector_private *p,
		struct wifi_network_device *dev)
{
	struct getter_context *gt =
		(struct getter_context *) dev->priv;

	if (gt && (gt->state >= GT_IDLE) && (gt->state < GT_DONE_1)) { /* cppcheck-suppress unsignedPositive */
		fprintf(stdout, "Getter for " MACFMT " is still running..... state = %s\n",
				MAC2STR(dev->macaddr), getter_state2str(gt->state));
		return 1;
	}

	return 0;
}

void getter_timer_cb(atimer_t *t)
{
	struct getter_context *gt = container_of(t, struct getter_context, t);
	struct decollector_private *p = (struct decollector_private *)gt->collector;
	struct wifi_network_device *dev = (struct wifi_network_device *)gt->agent;
	int ret;


	if (dev->invalid)
		return;

	switch (gt->request) {
	case CMDU_TYPE_TOPOLOGY_QUERY:
		if (!!(gt->rbitmap & TOPOLOGY_RESPONSE))
			return;
		break;

	case CMDU_AP_CAPABILITY_QUERY:
		if (!(gt->rbitmap & TOPOLOGY_RESPONSE)) {
			dev->invalid = true;
			return;
		}

		if (!!(gt->rbitmap & AP_CAP_REPORT))
			return;
		break;

	case CMDU_AP_METRICS_QUERY:
		if (!!(gt->rbitmap & AP_METRICS_RESPONSE))
			return;
		break;

	case CMDU_BACKHAUL_STA_CAPABILITY_QUERY:
		if (!!(gt->rbitmap & BSTA_CAP_REPORT))
			return;
		break;
	case CMDU_CLIENT_CAPABILITY_QUERY:
		if (!!(gt->rbitmap & STA_CAP_REPORT))
			return;
		break;

	case 0xffff:
	default:
		//TODO: decollector_stop_getter(p, dev);
		return;
	}

	fprintf(stderr, "decollector: Getter for %s to " MACFMT " --->\n",
		map_cmdu_type2str(gt->request), MAC2STR(dev->macaddr));

	ret = prepare_1905_query(p, gt->request, dev, dev->macaddr);
	if (!ret) {
		//FIXME
#if 0
		if (gt->request == CMDU_CLIENT_CAPABILITY_QUERY) {
			ret = decollector_send_client_cap_query(p, dev);

			/* TODO: check nothing else query in this refresh cycle */
			gt->retry = 0;
			gt->request = 0xffff;
			gt->tmo = 5000;
			timer_set(&gt->t, gt->tmo);
		}
		// dm->available = true ?
#endif

	} else {
		if (gt->retry++ < 1) {
			timer_set(&gt->t, 500);
			return;
		}

		if (!(gt->rbitmap & TOPOLOGY_RESPONSE))
			dev->invalid = true;
	}
}

int decollector_alloc_getter(struct decollector_private *p,
			     struct wifi_network_device *dev)
{
	struct getter_context *gt;

	gt = calloc(1, sizeof(struct getter_context));
	if (!gt)
		return -1;

	dev->priv = gt;
	timer_init(&gt->t, getter_timer_cb);
	gt->collector = p;
	gt->agent = dev;
	printf("Alloc getter timer %p  for " MACFMT "\n",
			&gt->t, MAC2STR(dev->macaddr));

	return 0;
}

int decollector_free_getter(struct wifi_network_device *dev)
{
	struct getter_context *gt;

	if (!dev->priv)
		return 0;

	gt = dev->priv;
	timer_del(&gt->t);
	gt->agent = NULL;
	gt->collector = NULL;

	free(gt);
	dev->priv = NULL;

	return 0;
}

int decollector_sched_getter(struct decollector_private *p,
			     struct wifi_network_device *dev,
			     uint16_t msg, uint32_t after_ms)
{
	struct getter_context *gt;

	//printf("Timer set for agent " MACFMT "\n", MAC2STR(dev->macaddr));
	gt = dev->priv;
	gt->request = msg;
	return timer_set(&gt->t, after_ms);
}

static void decolloctor_remove_not_found_devs(const uint8_t *discovered_macs,
					      int discovered_macs_number,
					      struct decollector_private *priv)
{
	struct wifi_network_device *dev = NULL, *tmp = NULL;

	list_for_each_entry_safe(dev, tmp, &priv->dm->network.devicelist, list) {
		int i;
		bool device_found = false;

		for (i = 0; i < discovered_macs_number; i++) {
			if (!memcmp(dev->macaddr, &discovered_macs[i * 6], 6)) {
				device_found = true;
				break;
			}
		}

		if (!device_found) {
			list_del(&dev->list);
			decollector_free_getter(dev);
			free_network_device(dev);
			priv->dm->network.num_devices--;
		}
	}
}

static uint8_t *decolloctor_create_list_of_new_devices(
	const uint8_t *discovered_macs, int discovered_macs_number,
	const struct decollector_private *priv, uint32_t *new_devs_number)
{
	struct wifi_network_device *dev;
	int i;
	uint32_t new_devices_number = 0;
	uint8_t *new_devices = calloc(discovered_macs_number, 6);

	if (!new_devices)
		return NULL;

	for (i = 0; i < discovered_macs_number; i++) {
		bool known_device = false;

		list_for_each_entry(dev, &priv->dm->network.devicelist, list) {
			if (!memcmp(dev->macaddr, &discovered_macs[i * 6], 6)) {
				known_device = true;
				break;
			}
		}

		if (!known_device) {
			memcpy(&new_devices[new_devices_number * 6],
			       &discovered_macs[i * 6], 6);
			++new_devices_number;
		}
	}

	*new_devs_number = new_devices_number;

	return new_devices;
}

static struct wifi_network_device *
decollector_alloc_network_device(const uint8_t *macaddr, struct decollector_private *priv)
{
	struct wifi_network_device *dev =
		calloc(1, sizeof(struct wifi_network_device));
	if (!dev) {
		err("%s: %d -ENOMEM\n", __func__, __LINE__);
		return NULL;
	}

	memcpy(dev->macaddr, macaddr, 6);
	if (decollector_alloc_getter(priv, dev)) {
		err("%s: %d -ENOMEM\n", __func__, __LINE__);
		free_network_device(dev);
		return NULL;
	}

	dev->map_profile = MAP_PROFILE_1;
	INIT_LIST_HEAD(&dev->radiolist);
	INIT_LIST_HEAD(&dev->cac_statuslist);
	INIT_LIST_HEAD(&dev->sta_steer_disallowlist);
	INIT_LIST_HEAD(&dev->sta_btmsteer_disallowlist);

	INIT_LIST_HEAD(&dev->ev.assoclist);
	INIT_LIST_HEAD(&dev->ev.disassoclist);
	INIT_LIST_HEAD(&dev->ev.failconnlist);

	return dev;

}

int decollector_alloc_deagents(struct decollector_private *priv)
{

	uint8_t *discovered_macs = NULL;
	int num_of_discovered = 0;
	uint8_t *new_devices = NULL;
	uint32_t new_devs_number = 0;
	struct wifi_network_device *dev = NULL;
	int ret;
	int i;

	ret = decollector_get_deagents(priv, &num_of_discovered, &discovered_macs);
	if (ret) {
		fprintf(stderr, "%s: ret = %d\n", __func__, ret);
		return ret;
	}

	dbg("%s: current devs number in model:  %d\n", __func__, priv->dm->network.num_devices);
	dbg("%s: discovered devs: %d\n", __func__, num_of_discovered);

	decolloctor_remove_not_found_devs(discovered_macs, num_of_discovered, priv);

	dbg("%s: devs number in model after removal:  %d\n", __func__, priv->dm->network.num_devices);

	new_devices = decolloctor_create_list_of_new_devices(
		discovered_macs, num_of_discovered, priv, &new_devs_number);

	dbg("%s: number of new devs to be added to model:  %d\n", __func__, new_devs_number);

	for (i = 0; i < new_devs_number; i++) {
		dev = decollector_alloc_network_device(&new_devices[i * 6], priv);

		dbg("%s: adding new dev: " MACFMT "\n", __func__, MAC2STR(&new_devices[i * 6]));

		if (dev) {
			list_add_tail(&dev->list, &priv->dm->network.devicelist);
			priv->dm->network.num_devices++;
		} else {
			ret = -1;
			goto out;
		}
	}

	info("%s: finally %d devs in model.\n", __func__, priv->dm->network.num_devices);
	list_for_each_entry(dev, &priv->dm->network.devicelist, list) {
		info("%s:   dev->macaddr: " MACFMT "\n", __func__, MAC2STR(dev->macaddr));
	}


out:
	free(discovered_macs);
	free(new_devices);
	return ret;
}

void decollector_getter_run(void *arg)
{
	struct decollector_private *p = (struct decollector_private *)arg;
	struct wifi_network_device *dev = NULL;
	struct wifi_data_element *dm = p->dm;
	int ret;


	fprintf(stderr, "decollector: Starting getter...\n");
	ret = collector_get_network_params(p, &dm->network);
	if (ret)
		err("%s: Error get network ssids!\n", __func__);

	p->cstate = CT_RUNNING;
	list_for_each_entry(dev, &dm->network.devicelist, list) {
		ret |= decollector_collect_node(p, dev);
	}

	remove_all_bh_topology_devices(&p->bh_topo_dev_list);

	ret |= decollector_get_network_steer_stats(p);
	ret |= decollector_get_sta_steer_stats(p);

	//TODO: handle ret
	fprintf(stderr, "decollector: Getter start done.\n");
}

void decollector_refresh_all(struct decollector_private *priv)
{
	time_t now;
	int ret;

	time(&now);
	if (difftime(now, priv->refresh_time) < 30) {
		fprintf(stderr, "%s: last refresh done within 30s\n", __func__);
		return;
	}

	ret = decollector_alloc_deagents(priv);
	if (ret)
		return;

	decollector_getter_run(priv);

	time(&priv->refresh_time);
}
