/*
 * decollector_ubus.c
 * WiFi Data Elements ubus methods
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <pthread.h>
#include <sys/types.h>
#include <unistd.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>

#include <easy/easy.h>
#include <wifiutils.h>


#include "timer.h"
#include <cmdu.h>
#include <1905_tlvs.h>
#include <easymesh.h>
#include <map_module.h>

#include "wifi_dataelements.h"
#include "decollector.h"
#include "debug.h"


#define UBUS_TIMEOUT		1000

extern int decollector_handle_map_event(struct decollector_private *p,
				struct cmdu_buff *cmdu);
extern int decollector_handle_topology_event(struct decollector_private *p,
				const char *type);


int decollector_dump(struct ubus_context *ctx, struct ubus_object *obj,
			struct ubus_request_data *req, const char *method,
			struct blob_attr *msg)
{
	struct decollector_private *p = container_of(obj, struct decollector_private, obj);
	struct blob_buf bb = {0};
	int ret;

	blob_buf_init(&bb, 0);

	ret = decollector_dump_all(p, &bb);
	if (ret) {
		fprintf(stderr, "DM is not ready; do a refresh!\n");
		goto out;
	}

	ubus_send_reply(ctx, req, bb.head);
out:
	blob_buf_free(&bb);

	return 0;
}

int decollector_dump2(struct ubus_context *ctx, struct ubus_object *obj,
	       struct ubus_request_data *req, const char *method,
	       struct blob_attr *msg)
{
	struct decollector_private *p = container_of(obj, struct decollector_private, obj);
	struct blob_buf bb;
	int ret = -1;

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);

	ret = decollector_dump2_all(p, &bb);
	if (ret) {
		fprintf(stderr, "DM is not ready; do a refresh!\n");
		goto out;
	}

	ubus_send_reply(ctx, req, bb.head);
out:
	blob_buf_free(&bb);

	return 0;
}

int decollector_refresh(struct ubus_context *ctx, struct ubus_object *obj,
			struct ubus_request_data *req, const char *method,
			struct blob_attr *msg)
{
	struct decollector_private *p =
		container_of(obj, struct decollector_private, obj);

	decollector_refresh_all(p);
	return 0;
}

int decollector_event(struct ubus_context *ctx, struct ubus_object *obj,
			struct ubus_request_data *req, const char *method,
			struct blob_attr *msg)
{
	struct blob_buf bb;
	struct decollector_private *p = container_of(obj, struct decollector_private, obj);

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);
	decollector_event_dump(p, &bb);
	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

int decollector_status(struct ubus_context *ctx, struct ubus_object *obj,
			struct ubus_request_data *req, const char *method,
			struct blob_attr *msg)
{
	struct blob_buf bb;
	struct decollector_private *p =
			container_of(obj, struct decollector_private, obj);

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "status", p->dm_available ? "available" : "incomplete");
	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

int decollector_publish_object(struct decollector_private *p, const char *objname)
{
	struct ubus_object *obj;
	struct ubus_object_type *obj_type;
	struct ubus_method *obj_methods;
	int ret;
	struct ubus_method m[5] = {
		UBUS_METHOD_NOARG("dump", decollector_dump),
		UBUS_METHOD_NOARG("dump2", decollector_dump2),
		UBUS_METHOD_NOARG("refresh", decollector_refresh),
		UBUS_METHOD_NOARG("event", decollector_event),
		UBUS_METHOD_NOARG("status", decollector_status),
	};
	int num_methods = ARRAY_SIZE(m);

	if (!p->ctx) {
		p->ctx = ubus_connect(NULL);
		if (!p->ctx) {
			fprintf(stderr, "Failed to connect to ubus\n");
			return -1;
		}
		ubus_add_uloop(p->ctx);
	}

	obj = &p->obj;
	memset(obj, 0, sizeof(*obj));

	obj_type = calloc(1, sizeof(struct ubus_object_type));
	if (!obj_type)
		return -1;

	obj_methods = calloc(num_methods, sizeof(struct ubus_method));
	if (!obj_methods) {
		free(obj_type);
		return -1;
	}

	obj->name = objname;
	memcpy(obj_methods, m, num_methods * sizeof(struct ubus_method));
	obj->methods = obj_methods;
	obj->n_methods = num_methods;

	obj_type->name = obj->name;
	obj_type->n_methods = obj->n_methods;
	obj_type->methods = obj->methods;
	obj->type = obj_type;

	ret = ubus_add_object(p->ctx, obj);
	if (ret) {
		fprintf(stderr, "Failed to add 'obj' err = %s\n",
			ubus_strerror(ret));
		free(obj_methods);
		free(obj_type);

		return ret;
	}

	fprintf(stderr, "ADDED object '%s' id %08x type id %08x\n", objname, obj->id, obj->type->id);

	return 0;
}

void decollector_ieee1905_cmdu_handler(struct decollector_private *priv,
		struct blob_attr *msg)
{
	uint16_t cmdu_type;
	uint16_t mid;
	char in_ifname[16] = {0};
	char src[18] = {0};
	uint8_t src_mac[6] = {0};
	uint8_t origin[6] = {0};
	int len = 0;
	char *tlvstr;
	uint8_t *tlv = NULL;
	struct cmdu_buff *cmdu;
	struct blob_attr *tb[6];
	static const struct blobmsg_policy cmdu_attr[6] = {
		[0] = { .name = "type", .type = BLOBMSG_TYPE_INT16 },
		[1] = { .name = "mid", .type = BLOBMSG_TYPE_INT16 },
		[2] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[3] = { .name = "src", .type = BLOBMSG_TYPE_STRING },
		[4] = { .name = "origin", .type = BLOBMSG_TYPE_STRING },
		[5] = { .name = "cmdu", .type = BLOBMSG_TYPE_STRING },
	};

	blobmsg_parse(cmdu_attr, 6, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1])
		return;

	if (tb[0]) {
		cmdu_type = (uint16_t)blobmsg_get_u16(tb[0]);
		if (cmdu_type > MAP_CMDU_TYPE_MAX)
			return;
	}

	if (tb[1])
		mid = (uint16_t)blobmsg_get_u16(tb[1]);

	if (tb[2])
		strncpy(in_ifname, blobmsg_data(tb[2]), 15);

	if (tb[3]) {
		strncpy(src, blobmsg_data(tb[3]), 17);
		hwaddr_aton(src, src_mac);
	}

	if (tb[4]) {
		memset(src, 0, sizeof(src));
		strncpy(src, blobmsg_data(tb[4]), 17);
		hwaddr_aton(src, origin);
	}

	if (tb[5]) {
		len = blobmsg_data_len(tb[5]) - 16;
		tlvstr = calloc(1, (len + 1) * sizeof(char));
		if (!tlvstr)
			return;

		strncpy(tlvstr, blobmsg_data(tb[5]) + 16, len);
		len = (len - 1) / 2;
		tlv = calloc(1, len * sizeof(uint8_t));
		if (!tlv) {
			free(tlvstr);
			return;
		}

		if (!strtob(tlvstr, len, tlv)) {
			free(tlvstr);
			free(tlv);
			return;
		}

		free(tlvstr);
	}

	cmdu = cmdu_alloc_custom(cmdu_type, &mid, in_ifname, src_mac, tlv, len);
	if (!cmdu) {
		fprintf(stderr, "%s: cmdu_alloc_custom failed!\n", __func__);
	} else {
		/* fprintf(stdout, "%s: cmdu type: %04x, mid: %hu, source: %s\n",
			__func__, cmdu_get_type(cmdu), cmdu_get_mid(cmdu), src);
		*/
		memcpy(cmdu->origin, origin, 6);
		decollector_handle_map_event(priv, cmdu);
		cmdu_free(cmdu);
	}

	if (tlv)
		free(tlv);
}

void decollector_topology_event_handler(struct ubus_context *ctx,
				struct ubus_event_handler *ev,
				const char *type,
				struct blob_attr *msg)
{
	enum {
		NEIGHBOR_AL_MAC,
		IS_1905_DEV,

		POLICY_SIZE
	};
	int is_1905;
	struct blob_attr *tb[POLICY_SIZE];

	static const struct blobmsg_policy neighbor_attrs[POLICY_SIZE] = {
		[NEIGHBOR_AL_MAC] = { .name = "nbr_ieee1905id", .type = BLOBMSG_TYPE_STRING },
		[IS_1905_DEV] = { .name = "is1905", .type = BLOBMSG_TYPE_INT8 },
	};
	struct decollector_private *priv =
			container_of(ev, struct decollector_private, topology_ev);


	blobmsg_parse(neighbor_attrs, POLICY_SIZE, tb, blob_data(msg), blob_len(msg));

	if (!tb[NEIGHBOR_AL_MAC])
		return;

	is_1905 = blobmsg_get_u8(tb[IS_1905_DEV]);
	if (!is_1905)
		return;

	decollector_handle_topology_event(priv, type);
}

void decollector_ubus_object_event_handler(struct ubus_context *ctx,
				struct ubus_event_handler *ev,
				const char *type,
				struct blob_attr *msg)
{
	char path[32] = { 0 };
	struct blob_attr *tb[2];
	static const struct blobmsg_policy ev_attr[2] = {
		[0] = { .name = "id", .type = BLOBMSG_TYPE_INT32 },
		[1] = { .name = "path", .type = BLOBMSG_TYPE_STRING }
	};
	struct decollector_private *p =
		container_of(ev, struct decollector_private, ubus_obj_ev);

	blobmsg_parse(ev_attr, 2, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1])
		return;

	strncpy(path, blobmsg_data(tb[1]), sizeof(path) - 1);

	if (strncmp(path, MAP_PLUGIN_OBJECT,
				strlen(MAP_PLUGIN_OBJECT)))
		return;

	/* update the i1905 & i1905.map object */
	p->i1905 = decollector_lookup_object(p, IEEE1905_OBJECT);
	if (p->i1905 == OBJECT_INVALID) {
		fprintf(stderr, "%s: Failed to get '%s' object!\n",
				__func__, IEEE1905_OBJECT);
		return;
	}

	p->i1905_map = decollector_lookup_object(p, MAP_PLUGIN_OBJECT);
	if (p->i1905_map == OBJECT_INVALID) {
		fprintf(stderr, "%s: Failed to get '%s' object!\n",
				__func__, MAP_PLUGIN_OBJECT);
		return;
	}

	fprintf(stderr, "re-subscribing CMDUs..!!\n");
	decollector_subscribe_for_cmdus(p);
}

int decollector_register_events(struct decollector_private *p)
{
	int ret;

	p->topology_ev.cb = decollector_topology_event_handler;
	ret = ubus_register_event_handler(p->ctx, &p->topology_ev, "ieee1905.neighbor.*");
	if (ret) {
		fprintf(stderr, "%s: failed %s\n for ieee1905.neighbor",
			__func__, ubus_strerror(ret));
	}

	p->ubus_obj_ev.cb = decollector_ubus_object_event_handler;
	ret = ubus_register_event_handler(p->ctx, &p->ubus_obj_ev, "ubus.object.add");
	if (ret) {
		fprintf(stderr, "%s: failed %s\n for ubus_object_add",
			__func__, ubus_strerror(ret));
	}

	return ret;
}

uint32_t decollector_lookup_object(struct decollector_private *p, const char *objname)
{
	uint32_t id;
	int status;

	status = ubus_lookup_id(p->ctx, objname, &id);
	if (status != UBUS_STATUS_OK) {
		fprintf(stderr, "object '%s' not present!\n", objname);
		return OBJECT_INVALID;
	}

	return id;
}

static void ieee1905_info_response_cb(struct ubus_request *req,
					int mtype, struct blob_attr *msg)
{
	struct json_object *jobj = NULL;
	struct json_object *mac;
	struct decollector_private *p;
	char *addr = NULL;
	char *str;

	if (!msg || !req->priv) {
		fprintf(stderr, "%s:Message recieved is NULL\n", __func__);
		return;
	}

	p = (struct decollector_private *)req->priv;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jobj = json_tokener_parse(str);
		free(str);
	}

	if (jobj == NULL)
		return;

	/* Store self device mac addr */
	if (json_object_object_get_ex(jobj, "ieee1905id", &mac)) {
		addr = strdup(json_object_get_string(mac));
		hwaddr_aton(addr, p->ieee1905_macaddr);
		free(addr);
	}

	json_object_put(jobj);
}

int decollector_get_ieee1905id(struct decollector_private *p)
{
	struct blob_buf bb = { 0 };
	int ret = 0;

	blob_buf_init(&bb, 0);
	ret = ubus_invoke(p->ctx, p->i1905, "info", bb.head,
				ieee1905_info_response_cb, p, 1000);
	if (ret)
		fprintf(stderr, "Failed to get i1905 info (err = %s)\n",
						ubus_strerror(ret));

	blob_buf_free(&bb);

	return ret;
}

static void steer_stats_response_cb(struct ubus_request *req, int mtype,
				    struct blob_attr *msg)
{
	enum {
		FAIL_NO_CANDIDATE,
		ASSOC_CNTRL_ATTEMPTS,
		ASSOC_CNTRL_SUCCESS,
		ASSOC_CNTRL_FAIL,
		BTM_ATTEMPTS,
		BTM_SUCCESS,
		BTM_FAIL,
		BTM_QUERY_RESP,
		BTM_TIME_SINCE_STEER_ATTEMPT,

		STEER_RESP_NUM_PARAMS
	};

	static const struct blobmsg_policy resp_policy[STEER_RESP_NUM_PARAMS] = {
		[FAIL_NO_CANDIDATE] = { .name = "fail_no_candidate", .type = BLOBMSG_TYPE_INT64 },
		[ASSOC_CNTRL_ATTEMPTS] = { .name = "assoc_cntlr_attempts", .type = BLOBMSG_TYPE_INT64 },
		[ASSOC_CNTRL_SUCCESS] = { .name = "assoc_cntlr_success", .type = BLOBMSG_TYPE_INT64 },
		[ASSOC_CNTRL_FAIL] = { .name = "assoc_cntlr_fail", .type = BLOBMSG_TYPE_INT64 },
		[BTM_ATTEMPTS] = { .name = "btm_attempts", .type = BLOBMSG_TYPE_INT64 },
		[BTM_SUCCESS] = { .name = "btm_success", .type = BLOBMSG_TYPE_INT64 },
		[BTM_FAIL] = { .name = "btm_fail", .type = BLOBMSG_TYPE_INT64 },
		[BTM_QUERY_RESP] = { .name = "btm_query_resp", .type = BLOBMSG_TYPE_INT64 },
		[BTM_TIME_SINCE_STEER_ATTEMPT] =
			{ .name = "time_since_steer_attempt", .type = BLOBMSG_TYPE_INT32 },
	};

	struct wifi_steer_summary *dm_steer_stats;
	struct blob_attr *tb[STEER_RESP_NUM_PARAMS] = { NULL };

	if (!msg || !req->priv) {
		fprintf(stderr, "%s:Message recieved is NULL\n", __func__);
		return;
	}

	fprintf(stderr, "%s: collecting steer_stats...\n", __func__);

	dm_steer_stats = (struct wifi_steer_summary *)req->priv;
	*dm_steer_stats = (struct wifi_steer_summary){ 0 };

	blobmsg_parse(resp_policy, STEER_RESP_NUM_PARAMS, tb, blob_data(msg), blob_len(msg));

	if (tb[FAIL_NO_CANDIDATE]) {
		dm_steer_stats->no_candidate_cnt =
			blobmsg_get_u64(tb[FAIL_NO_CANDIDATE]);
	}

	if (tb[ASSOC_CNTRL_ATTEMPTS]) {
		dm_steer_stats->blacklist_attempt_cnt =
			blobmsg_get_u64(tb[ASSOC_CNTRL_ATTEMPTS]);
	}

	if (tb[ASSOC_CNTRL_SUCCESS]) {
		dm_steer_stats->blacklist_success_cnt =
			blobmsg_get_u64(tb[ASSOC_CNTRL_SUCCESS]);
	}

	if (tb[ASSOC_CNTRL_FAIL]) {
		dm_steer_stats->blacklist_failure_cnt =
			blobmsg_get_u64(tb[ASSOC_CNTRL_FAIL]);
	}

	if (tb[BTM_ATTEMPTS]) {
		dm_steer_stats->btm_attempt_cnt =
			blobmsg_get_u64(tb[BTM_ATTEMPTS]);
	}

	if (tb[BTM_SUCCESS]) {
		dm_steer_stats->btm_success_cnt =
			blobmsg_get_u64(tb[BTM_SUCCESS]);
	}

	if (tb[BTM_FAIL]) {
		dm_steer_stats->btm_failure_cnt =
			blobmsg_get_u64(tb[BTM_FAIL]);
	}

	if (tb[BTM_QUERY_RESP]) {
		dm_steer_stats->btm_query_resp_cnt =
			blobmsg_get_u64(tb[BTM_QUERY_RESP]);
	}

	if (tb[BTM_TIME_SINCE_STEER_ATTEMPT]) {
		dm_steer_stats->last_tsp.tv_sec =
			blobmsg_get_u32(tb[BTM_TIME_SINCE_STEER_ATTEMPT]);
	}
}

static int decollector_get_steer_stats_impl(struct decollector_private *p,
					    struct wifi_steer_summary *dm_steer_stats,
					    const char *sta_macaddr_str)
{
	const int ONE_SEC_TIMEOUT = 1000;
	uint32_t id;
	int ret;
	struct blob_buf msg = { 0 };

	id = decollector_lookup_object(p, MAP_CONTROLLER_OBJECT);
	if (id == OBJECT_INVALID)
		return -1;

	blob_buf_init(&msg, 0);

	if (sta_macaddr_str)
		blobmsg_add_string(&msg, "sta", sta_macaddr_str);

	ret = ubus_invoke(p->ctx, id, "steer_summary", msg.head,
			  steer_stats_response_cb, dm_steer_stats, ONE_SEC_TIMEOUT);

	if (ret) {
		fprintf(stderr, "Failed to get steer_stats for %s (err = %s)\n",
			sta_macaddr_str ? sta_macaddr_str : "network",
			ubus_strerror(ret));
	}

	blob_buf_free(&msg);

	return ret;
}

int decollector_get_network_steer_stats(struct decollector_private *p)
{
	return decollector_get_steer_stats_impl(p,
						&p->dm->network.steer_summary,
						NULL);
}

int decollector_get_sta_steer_stats(struct decollector_private *p)
{
	struct wifi_network *net = &p->dm->network;
	struct wifi_network_device *device = NULL;
	struct wifi_radio_element *radio = NULL;
	struct wifi_bss_element *bss = NULL;
	struct wifi_sta_element *sta = NULL;
	int res = 0;

	list_for_each_entry(device, &net->devicelist, list) {
		list_for_each_entry(radio, &device->radiolist, list) {
			list_for_each_entry(bss, &radio->bsslist, list) {
				list_for_each_entry(sta, &bss->stalist, list) {
					char sta_macaddr_str[18] = { 0 };
					struct wifi_steer_summary *out =
						&sta->mapsta.stats;

					if (!hwaddr_ntoa(sta->macaddr,
							 sta_macaddr_str)) {
						fprintf(stderr,
							"%s: sta_macaddr is not valid)\n",
							__func__);
						continue;
					}

					res |= decollector_get_steer_stats_impl(
						p, out, sta_macaddr_str);
				}
			}
		}
	}

	return res;
}

static void decollector_send_cmdu_cb(struct ubus_request *req, int type,
				     struct blob_attr *msg)
{
	struct json_object *jobj = NULL;
	struct json_object *tmp;
	uint16_t *mid;
	char *str;

	if (!msg || !req->priv) {
		fprintf(stderr, "%s:Message recieved is NULL\n", __func__);
		return;
	}

	mid = (uint16_t *)req->priv;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jobj = json_tokener_parse(str);
		free(str);
	}

	if (jobj == NULL)
		return;

	if (json_object_object_get_ex(jobj, "mid", &tmp))
		*mid = json_object_get_int(tmp);


	json_object_put(jobj);
}

int decollector_send_cmdu_request(struct decollector_private *priv,
				  struct cmdu_buff *req, uint8_t *dst)
{
	//struct wifi_network_device *dev = NULL;
	uint16_t type = cmdu_get_type(req);
	struct blob_buf b = { 0 };
	char dst_addr[18] = { 0 };
	uint16_t msgid = 0;
	int ret = 0;


	fprintf(stderr, "decollector: sending %s --->\n", map_cmdu_type2str(type));

	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);
	blobmsg_add_u32(&b, "type", cmdu_get_type(req));
	blobmsg_add_u32(&b, "mid", cmdu_get_mid(req));
	hwaddr_ntoa(dst, dst_addr);
	blobmsg_add_string(&b, "dst", dst_addr);

	if (req->datalen) {
		char *tlv_str = NULL;
		uint16_t len = 0;

		len = (req->datalen * 2) + 1;

		tlv_str = calloc(len, sizeof(char));
		if (!tlv_str)
			goto out;

		btostr(req->data, req->datalen, tlv_str);
		tlv_str[len-1] = '\0';

		blobmsg_add_string(&b, "data", tlv_str);
		free(tlv_str);
	}

	ret = ubus_invoke(priv->ctx, priv->i1905, "cmdu",
			  b.head, decollector_send_cmdu_cb,
			  (void *)&msgid,
			  UBUS_TIMEOUT);
	if (!ret) {
		fprintf(stderr, "%s: mid = %hu\n", __func__, msgid);
	} else {
		fprintf(stderr, "%s: cmdu send failed (err = %s)\n",
			__func__, ubus_strerror(ret));

#if 0
		dev = decollector_get_origin_dev(priv, priv->dm, req->origin);
		if (dev)
			decollector_stop_getter(priv, dev);

		if (ret == UBUS_STATUS_NOT_FOUND) {
			/* refresh reference to ieee1905 object */
			priv->i1905 = decollector_lookup_object(priv, IEEE1905_OBJECT);
			if (priv->i1905 == OBJECT_INVALID)
				fprintf(stderr, "Failed to get '%s' object!\n", IEEE1905_OBJECT);
		}
#endif
	}

out:
	blob_buf_free(&b);
	return ret;
}

static void ieee1905_neighbors_cb(struct ubus_request *req,
		int mtype, struct blob_attr *msg)
{
	int i, num;
	int total_nodes = 0;
	struct json_object *nbr_array;
	struct json_object *jobj = NULL;
	char *str = NULL;
	struct map_devices {
		int num;
		uint8_t *dev;
	} *n_array = NULL;


	if (!msg || !req->priv) {
		fprintf(stderr, "%s: msg = NULL\n", __func__);
		return;
	}

	n_array = (struct map_devices *)req->priv;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jobj = json_tokener_parse(str);
		free(str);
	}

	if (jobj == NULL)
		return;

	/* for own 1905 macaddress */
	total_nodes += 1;

	if (json_object_object_get_ex(jobj, "neighbors", &nbr_array)) {
		num = json_object_array_length(nbr_array);
		if (!num)
			fprintf(stderr, "%s: nodes array is empty\n", __func__);

		total_nodes += num;

		n_array->dev = calloc(total_nodes, 6 * sizeof(uint8_t));
		if (!n_array->dev) {
			fprintf(stderr, "Failed to malloc nodes array\n");
			goto out_err;
		}

		n_array->num = total_nodes;

		for (i = 0; i < num; i++) {

			struct json_object *nbr_object
				= json_object_array_get_idx(nbr_array, i);
			struct json_object *mac_addr_object = NULL;

			if (json_object_object_get_ex(nbr_object, "ieee1905id", &mac_addr_object)) {
				const char *mac_addr_str =
					json_object_get_string(mac_addr_object);

				if (mac_addr_str) {
					uint8_t tmpmac[6] = { 0 };
					hwaddr_aton(mac_addr_str, tmpmac);
					memcpy(&n_array->dev[6 * (i + 1)], tmpmac, 6);
				}
			} else {
				fprintf(stderr, "%s: Failed to get ieee1905id object for device[%d]\n", __func__, i);
			}
		}
	}


out_err:
	json_object_put(jobj);
}

int decollector_get_deagents(struct decollector_private *p, int *num, uint8_t **macaddrs)
{
	struct blob_buf bb;
	int ret = 0;
	struct map_devices {
		int num;
		uint8_t *dev;
	} mapdevs = { 0 };


	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	ret = ubus_invoke(p->ctx, p->i1905, "neighbors", bb.head,
			  ieee1905_neighbors_cb, &mapdevs, 1000);

	if (!ret) {
		/* copy the self device macaddr */
		memcpy(&mapdevs.dev[0], p->ieee1905_macaddr, 6);

		*num = mapdevs.num;
		*macaddrs = mapdevs.dev;
		fprintf(stderr, "num ieee1905 nodes = %d\n", *num);
	} else {
		fprintf(stderr, "Failed to call ubus ieee1905 neighbors\n");
	}

	blob_buf_free(&bb);

	return ret;
}

int decollector_map_sub_cb(void *bus, void *priv, void *data)
{
	struct blob_attr *msg = (struct blob_attr *)data;

/*
	char *str;

	str = blobmsg_format_json(msg, true);
	fprintf(stdout, "decollector: Received: %s\n", str);
	free(str);
*/

	decollector_ieee1905_cmdu_handler(priv, msg);
	return 0;
}

int decollector_map_del_cb(void *bus, void *priv, void *data)
{
	uint32_t *obj = (uint32_t *)data;

	fprintf(stdout, "Object 0x%x no longer present\n", *obj);
	return 0;
}

int decollector_subscribe_for_cmdus(struct decollector_private *priv)
{
	mapmodule_cmdu_mask_t cmdu_mask = {0};
	int ret;

	map_prepare_cmdu_mask(cmdu_mask,
			CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE,
			CMDU_TYPE_TOPOLOGY_RESPONSE,
			CMDU_AP_CAPABILITY_REPORT,
			CMDU_AP_METRICS_RESPONSE,
			CMDU_UNASSOC_STA_LINK_METRIC_RESPONSE,
			CMDU_BACKHAUL_STA_CAPABILITY_REPORT,
			CMDU_OPERATING_CHANNEL_REPORT,
			CMDU_CHANNEL_SCAN_REPORT,
			CMDU_CHANNEL_PREFERENCE_REPORT,
			CMDU_TYPE_TOPOLOGY_NOTIFICATION,
			CMDU_CLIENT_DISASSOCIATION_STATS,
			CMDU_CLIENT_CAPABILITY_REPORT,
			CMDU_TYPE_INVALID);

	ret = map_subscribe(priv->ctx,
			    &priv->i1905_map,
			    "decollector", &cmdu_mask, priv,
			    decollector_map_sub_cb,
			    decollector_map_del_cb,
			    &priv->subscriber);
	return ret;
}
